program marCCD("P=marCCD:,C=ccd1")

/*  This sequencer program works with ccd.db.  
 *  It allows EPICS control of the MAR 165 CCD detector.  The detector must 
 *  be put in remote mode, running the marccd_server_socket program.
 *
 *  NOTE: I am using "sprintf" where I would like to use "snprintf", but
 *        vxWorks does not have snprintf.
 *
 *  Mark Rivers
 *  Sept 27, 2003
*/

/* This program must be compiled with the recursive option so that 
 * multiple copies can run at once
 */
option +r;

%%#include <string.h>
%%#include <stdio.h>
%%#include <errlog.h>
%%#include <epicsTime.h>
%%#include <epicsEvent.h>
%%#include <asynDriver.h>
%%#include <asynSyncIO.h>
#include "epicsCCD.h"
#define MAX_COMMAND_SIZE 200
#define TIMEOUT 1.0 /* Default timeout */

/* Define escaped C functions at end of file */
%% static int asynConnect(SS_ID ssId, struct UserVar *pVar);
%% static int asynWrite(SS_ID ssId, struct UserVar *pVar, char *command);
%% static int asynWriteRead(SS_ID ssId, struct UserVar *pVar, char *command, 
%%                          double timeout);

/* Variables */
char   buffer[MAX_COMMAND_SIZE];
double exposureTime;
double delay;
char   *blank_pos;
char   *portName;
int    nchars;
int    save_single;
int    status;
int    validPath;
int    AsynPortIndex;
int    DetOutStrIndex;
int    DetInStrIndex;
/* Note, have to fake out the following definitions, because we need these
 * variables to be put into the UserVar structure (for +r option), but SNL
 * only knows about char*, not these more complex things */
char   *startTime;    /* Really epicsTimeStamp* */
char   *currentTime;  /* Really epicsTimeStamp* */
char   *abortEvent;   /* Really epicsEventId */
char   *pasynUser;    /* Really asynUser */

ss main {
    state init {
        when() {
            efClear(SetTemp_mon);
            efClear(AcquireCLBK_mon);
            efClear(AcquirePOLL_mon);
            efClear(Abort_mon);
            efClear(SaveFile_mon);
            efClear(ServerName_mon);
            efClear(ServerPort_mon);
            strncpy(Comment1Desc, "Det. Dist. (mm)", sizeof(Comment1Desc));  
            pvPut(Comment1Desc);
            strncpy(Comment2Desc, "Rotation axis", sizeof(Comment2Desc));
            pvPut(Comment2Desc);
            strncpy(Comment3Desc, "Wavelength", sizeof(Comment3Desc));
            pvPut(Comment3Desc);
            strncpy(Comment4Desc, "File comment", sizeof(Comment4Desc));
            pvPut(Comment4Desc);
            strncpy(Comment5Desc, "Dataset comment", sizeof(Comment5Desc));
            pvPut(Comment5Desc);
            efSet(ShutterStatLink_mon); /* Copy link first time through */
            ConnectState = STATE_DISCONNECTED; 
            pvPut(ConnectState);
            Initialize = 1; pvPut(Initialize); /* Force initialize first time */
            strncpy(CCDManufacturer, "MAR", sizeof(CCDManufacturer)); 
            pvPut(CCDManufacturer);
            strncpy(CCDModel, "165", sizeof(CCDModel)); 
            pvPut(CCDModel);
            MeasuredTemp = 25; 
            pvPut(MeasuredTemp);
            BitDepth = 16; 
            pvPut(BitDepth);
            AsynPortIndex = pvIndex(AsynPort);
            DetInStrIndex = pvIndex(DetInStr);
            DetOutStrIndex = pvIndex(DetOutStr);
            %%pVar->abortEvent = (char *) epicsEventCreate(epicsEventEmpty);
            /* Allocate epicsTimeStamp objects */
            %%pVar->startTime   = (char *) calloc(1, sizeof(epicsTimeStamp));
            %%pVar->currentTime = (char *) calloc(1, sizeof(epicsTimeStamp));
        } state wait_connect
    }

    state wait_connect {
        entry {
            ConnectState = STATE_DISCONNECTED;
            pvPut(ConnectState);
        }
        when((ConnectState == STATE_DISCONNECTED) &&
             ((efTestAndClear(Initialize_mon) && Initialize == 1) ||
              (efTestAndClear(ServerName_mon)) ||
              (efTestAndClear(ServerPort_mon)))) {
            /* Set the PV back to 0 */
            Initialize = 0; pvPut(Initialize);
            /* Try to create socket connection to server */
            %%pVar->status = asynConnect(ssId, pVar);
            if (status == 0) ConnectState = STATE_CONNECTED;
            else ConnectState = STATE_DISCONNECTED;
            pvPut(ConnectState);
        } state wait_connect

        when(ConnectState == STATE_CONNECTED) {
        } state monitor_changes
 
    }

    state monitor_changes {
        when((ConnectState == STATE_DISCONNECTED) ||
             (efTest(Initialize_mon) && Initialize == 1) ||
             (efTest(ServerName_mon)) ||
             (efTest(ServerPort_mon))) {
        } state wait_connect

        when(efTestAndClear(PollDetState_mon)) { 
            %%asynWriteRead(ssId, pVar, "get_state\n", TIMEOUT);
            DetectorState = atoi(buffer); pvPut(DetectorState);
            %%asynWriteRead(ssId, pVar, "get_bin\n", TIMEOUT);
            sscanf(buffer, "%d,%d", &ActualBinX, &ActualBinY);
            pvPut(ActualBinX); pvPut(ActualBinY);
        } state monitor_changes

        when(efTestAndClear(ShutterStatLink_mon)) {
            /* Copy the link to the shutter PV, but strip off CP MS*/
            blank_pos = strchr(ShutterStatLink, ' ');
            if (blank_pos != 0) nchars = blank_pos-ShutterStatLink; 
            else nchars=strlen(ShutterStatLink);
            strncpy(buffer, ShutterStatLink, sizeof(buffer));
            buffer[nchars] = 0;
            pvAssign(Shutter, buffer);
            /* Process the shutter status PV so the link status will change */
            /* Need to wait a little while for link to establish before 
            /* processing status record */
            epicsThreadSleep(0.1);
            ShutterStatProc = 1; pvPut(ShutterStatProc);
        } state monitor_changes

        when(efTestAndClear(OpenShutter_mon) && 
                            OpenShutter==1) {
            /* Are we using an EPICS PV for the shutter, or the MAR 165 
             * shutter? */
            if (ShutterMode == SHUTTER_EPICS) { 
                strncpy(Shutter, OpenShutterStr, sizeof(Shutter)); 
                pvPut(Shutter);
            } else if (ShutterMode == SHUTTER_CAMERA) {
                %%asynWrite(ssId, pVar, "shutter,1\n");
            }
            /* Set the PV back to 0 */
            OpenShutter = 0; pvPut(OpenShutter);
        } state monitor_changes

        when(efTestAndClear(CloseShutter_mon) && 
                            CloseShutter==1) {
            /* Are we using an EPICS PV for the shutter, or the MAR 165 
             * shutter? */
            if (ShutterMode == SHUTTER_EPICS) { 
                strncpy(Shutter, CloseShutterStr, sizeof(Shutter)); 
                pvPut(Shutter);
            } else if (ShutterMode == SHUTTER_CAMERA) {
                %%asynWrite(ssId, pVar, "shutter,0\n");
            }
            /* Set the PV back to 0 */
            CloseShutter = 0; pvPut(CloseShutter);
        } state monitor_changes

        when((efTestAndClear(BinX_mon)) ||
             (efTestAndClear(BinY_mon))) {
            /* Set binning */
            sprintf(buffer, "set_bin,%d,%d\n", BinX, BinY);
            %%asynWrite(ssId, pVar, pVar->buffer);
        } state monitor_changes

        when(efTestAndClear(FilePath_mon)) {
            /* Make sure path has a trailing / */
            if ((FilePath[strlen(FilePath)-1] != '/') &&
                (strlen(FilePath) < sizeof(FilePath)-2)) {
               strcat(FilePath, "/");
               pvPut(FilePath);
            }
            /* Check that this is a valid path */
/*           validPath = 1;
           %%if (opendir(pVar->FilePath) == NULL) pVar->validPath = 0;
           if (validPath == 0) {
               strcpy(FilePath, "./");
               pvPut(FilePath);
            }
*/
        } state monitor_changes

        when(efTestAndClear(SaveFile_mon) && 
                            SaveFile==1) {
            /* Set the PV back to 0 */
            SaveFile = 0; pvPut(SaveFile);
            /* Save file */
            save_single = 1;
        } state save_file

        when((efTestAndClear(AcquireCLBK_mon) && AcquireCLBK==1) ||
             (efTestAndClear(AcquirePOLL_mon) && AcquirePOLL==1)) {
        } state acquire
    }

    state acquire {
        when(FrameType == FRAME_BACKGROUND) {
        } state acquire_background

        when(FrameType != FRAME_BACKGROUND) {
        } state acquire_normal
    }

    state acquire_background {
        when () {
            /* Acquire 2 backgrounds with 1msec exposure each.  First into
             * scratch buffer, second into background buffer, then dezinger
             * Minimum exposure time of 1msec, because the readout command 
             * is not seen if it comes after the start command with no delay */
            exposureTime = .001;
            /* Acquire first background into scratch buffer */
            %%asynWrite(ssId, pVar, "start\n");
            while(DetectorState != DETECTOR_ACQUIRE) {
               epicsThreadSleep(.05);
               /* Read the status */
               %%asynWriteRead(ssId, pVar, "get_state\n", TIMEOUT);
               DetectorState = atoi(buffer); pvPut(DetectorState);
            }
            epicsThreadSleep(exposureTime);
            /* Start readout into scratch buffer*/
            /* Detector sends 0 when readout is done */
            DetectorState = DETECTOR_READOUT; pvPut(DetectorState);
            %%asynWriteRead(ssId, pVar, "readout,2\n", -1.);
            DetectorState = atoi(buffer); pvPut(DetectorState);
            while (DetectorState != DETECTOR_IDLE) {
               epicsThreadSleep(0.05);
               /* Read the status */
               %%asynWriteRead(ssId, pVar, "get_state\n", TIMEOUT);
               DetectorState = atoi(buffer); pvPut(DetectorState);
            }

            /* Acquire second background into background buffer */ 
            %%asynWrite(ssId, pVar, "start\n");
            while(DetectorState != DETECTOR_ACQUIRE) {
               epicsThreadSleep(.05);
               /* Read the status */
               %%asynWriteRead(ssId, pVar, "get_state\n", TIMEOUT);
               DetectorState = atoi(buffer); pvPut(DetectorState);
            }
            epicsThreadSleep(exposureTime);
            /* Start readout into background buffer*/
            /* Detector sends 0 when readout is done */
            DetectorState = DETECTOR_READOUT; pvPut(DetectorState);
            %%asynWriteRead(ssId, pVar, "readout,1\n", -1.);
            DetectorState = atoi(buffer); pvPut(DetectorState);
            while (DetectorState != DETECTOR_IDLE) {
               epicsThreadSleep(0.05);
               /* Read the status */
               %%asynWriteRead(ssId, pVar, "get_state\n", TIMEOUT);
               DetectorState = atoi(buffer); pvPut(DetectorState);
            }

            /* Remove zingers from background */
            %%asynWrite(ssId, pVar, "dezinger,1\n");
        } state save_file
    }


    state acquire_normal {
        when() {
            /* Start acquisition */
            exposureTime = Hours*3600. + Minutes*60. + Seconds +
                           Milliseconds/1000.;
            /* Minimum exposure time of 1msec, because the readout command
             * is not seen if it comes after the start command with no delay */
            if (exposureTime < .001) exposureTime = .001;
            %%asynWrite(ssId, pVar, "start\n");
            while (DetectorState != DETECTOR_ACQUIRE) {
               epicsThreadSleep(0.05);
               /* Read the status */
               %%asynWriteRead(ssId, pVar, "get_state\n", TIMEOUT);
               DetectorState = atoi(buffer); pvPut(DetectorState);
            }

            /* Set the the start time for the TimeRemaining counter */
            %%epicsTimeGetCurrent((epicsTimeStamp *)pVar->startTime);
            TimeRemaining = exposureTime; pvPut(TimeRemaining);
            /* Open the shutter */
            /* EPICS PV for the shutter, or the MAR 165 shutter? */
            if (ShutterMode == SHUTTER_EPICS) { 
                strncpy(Shutter, OpenShutterStr, sizeof(Shutter)); 
                pvPut(Shutter);
            } else if (ShutterMode == SHUTTER_CAMERA) {
                %%asynWrite(ssId, pVar, "shutter,1\n");
            }
            /* This delay is to get the exposure time correct.  
             * It is equal to the opening time of the shutter minus the
             * closing time.  If they are equal then no delay is needed, 
             * except use 1msec so delay is not negative and commands are 
             * not back-to-back */
            delay = OpenShutterDly-CloseShutterDly;
            if (delay < .001) delay=.001;
            epicsThreadSleep(delay);
            /* Wait for the exposure time using epicsEventWaitWithTimeout, 
             * so we can abort */
            %%epicsEventWaitWithTimeout((epicsEventId)pVar->abortEvent, 
            %%                          pVar->exposureTime);
            /* Close shutter */
            /* Are we using an EPICS PV for the shutter, or the 
             * MAR 165 shutter? */
            if (ShutterMode == SHUTTER_EPICS) { 
                strncpy(Shutter, CloseShutterStr, sizeof(Shutter)); 
                pvPut(Shutter);
            } else if (ShutterMode == SHUTTER_CAMERA) {
                %%asynWrite(ssId, pVar, "shutter,0\n");
            }
            epicsThreadSleep(CloseShutterDly);
            /* Start readout */
            /* Detector sends 0 when readout is done */
            DetectorState = DETECTOR_READOUT; pvPut(DetectorState);
            if (FrameType == FRAME_BACKGROUND) 
                %%asynWriteRead(ssId, pVar, "readout,1\n", -1.);
            else                               
                 %%asynWriteRead(ssId, pVar, "readout,0\n", -1.);
            DetectorState = atoi(buffer); pvPut(DetectorState);

            while (DetectorState != DETECTOR_IDLE) {
                epicsThreadSleep(0.05);
                /* Read the status */
                %%asynWriteRead(ssId, pVar, "get_state\n", TIMEOUT);
                DetectorState = atoi(buffer); pvPut(DetectorState);
            }

        } state save_file
    }

    state save_file {
        when() {
            if (AutoSave || save_single) {
                /* Construct header command */
                sprintf(buffer,  
                        "header,detector_distance=%s,exposure_time=%f,"
                        "rotation_axis=%s\n",
                        Comment1, exposureTime, Comment2);
                %%asynWrite(ssId, pVar, pVar->buffer);
                sprintf(buffer, 
                        "header,source_wavelength=%s,file_comments=%s,"
                        "dataset_comments=%s\n",
                        Comment3, Comment4, Comment5);
                %%asynWrite(ssId, pVar, pVar->buffer);
                if (CorrectFlatfield) {
                    DetectorState = DETECTOR_CORRECT; pvPut(DetectorState);
                    /* Without the following 20msec delay we get an access 
                     * fault in listen_socket().*/
                    epicsThreadSleep(.02);
                    /* Server sends back 0 when the correction is complete */
                    %%asynWriteRead(ssId, pVar, "correct", 10.0);
                }
                /* Save file */
                sprintf(FullFilename, FilenameFormat, FileTemplate, SeqNumber);
                pvPut(FullFilename);
                SeqNumber++; pvPut(SeqNumber);
                sprintf(buffer, "writefile,%s%s,%d\n", 
                        FilePath, FullFilename, CorrectFlatfield?1:0);
                DetectorState = DETECTOR_WRITING; pvPut(DetectorState);
                /* Server sends back 0 when the write is complete */
                %%asynWriteRead(ssId, pVar, pVar->buffer, -1.0);
                save_single = 0;
            }
            while (DetectorState != DETECTOR_IDLE) {
               epicsThreadSleep(0.05);
               /* Read the status */
               %%asynWriteRead(ssId, pVar, "get_state\n", TIMEOUT);
               DetectorState = atoi(buffer); pvPut(DetectorState);
            }
            AcquireCLBK = 0; pvPut(AcquireCLBK);
            AcquirePOLL = 0; pvPut(AcquirePOLL);
        } state monitor_changes
    }
}

/* This is a separate state set so the watchdog runs even during long 
 * acquisitions */
ss watchdog {
    state poll {
        when(delay(1.0) && SNLWatchdog < 5) {
            SNLWatchdog = 5; pvPut(SNLWatchdog);
        } state poll
    }
}

/* This is a separate state set so we can abort during the 
 * epicsEventWaitWithTimeout that times the shutter */
ss abort {
    state poll {
        when(Abort_mon && Abort==1) {
            %%asynWrite(ssId, pVar, "abort");
            %%epicsEventSignal((epicsEventId) pVar->abortEvent);
            Abort=0; pvPut(Abort);
            TimeRemaining=0.; pvPut(TimeRemaining);
        } state poll
    }
}

/* This is a separate state set so we update the time remaining during the 
 * epicsEventWaitWithTimeout that times the shutter */
ss timeRemaining {
    state poll {
        when(delay(.1) && TimeRemaining != 0.) {
            %%epicsTimeGetCurrent((epicsTimeStamp *) pVar->currentTime);
            %%pVar->TimeRemaining = pVar->exposureTime - 
            %%      epicsTimeDiffInSeconds(
            %%           (epicsTimeStamp *)pVar->currentTime, 
            %%           (epicsTimeStamp *)pVar->startTime);
            if (TimeRemaining < 0.) TimeRemaining = 0.;
            pvPut(TimeRemaining);
        } state poll
    }
}

/* Begin C code functions */
%{

static int asynConnect(SS_ID ssId, struct UserVar *pVar)
{   
   asynStatus status;
   char *portString;

   status = pasynSyncIO->openSocket(pVar->ServerName, pVar->ServerPort, 
                                    &portString);
   if (status != asynSuccess) return(status);
   status = pasynSyncIO->connect(portString, 0, (asynUser **)&pVar->pasynUser);
   if (status != asynSuccess) return(status);
   pasynTrace->setTraceIOMask((asynUser *)pVar->pasynUser, ASYN_TRACEIO_ESCAPE);
   /* Also connect the asynRecord for user interaction */
   strncpy(pVar->AsynPort, portString, sizeof(pVar->AsynPort));
   seq_pvPut(ssId, pVar->AsynPortIndex, 0);
   return(status);
}

static int asynWrite(SS_ID ssId, struct UserVar *pVar, char *command)
{
   int nwrite;

   if (pVar->ConnectState != STATE_CONNECTED) return(-1);
   /* A short delay is required between sending commands to MAR, else they 
    * get partly lost? */
   epicsThreadSleep(.001);
   nwrite = pasynSyncIO->write((asynUser *)pVar->pasynUser, command, 
                               strlen(command), TIMEOUT);
   /* Copy output string to PV so user can see it */
   strncpy(pVar->DetOutStr, command, sizeof(pVar->DetOutStr));
   seq_pvPut(ssId, pVar->DetOutStrIndex, 0);
   return(nwrite);
}

static int asynWriteRead(SS_ID ssId, struct UserVar *pVar, char *command,
                         double timeout)
{
   int nread;

   if (pVar->ConnectState != STATE_CONNECTED) return(-1);
   /* A short delay is required between sending commands to MAR, else they
    * get partly lost */
   epicsThreadSleep(.001);
   nread = pasynSyncIO->writeRead((asynUser *)pVar->pasynUser, command, 
                                  strlen(command), pVar->buffer, 
                                  sizeof(pVar->buffer), "\n", 1, timeout);

   /* Copy output string to PV so user can see it */
   strncpy(pVar->DetOutStr, command, sizeof(pVar->DetOutStr));
   seq_pvPut(ssId, pVar->DetOutStrIndex, 0);
   /* Copy input string to PV so user can see it */
   strncpy(pVar->DetInStr, pVar->buffer, sizeof(pVar->DetInStr));
   seq_pvPut(ssId, pVar->DetInStrIndex, 0);
   return(nread);
}

}%
